/*
 * Copyright (C) 1997 Steve Ginzburg.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#pragma ident	"@(#)devmap.c	1.4	99/02/23 SMI"


/*
**  This code maps major device numbers to the appropriate maps from
**  Linux ioctl()'s to native ioctl()'s.
*/


#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/select.h>
#if defined(UNIXWARE) || defined(GEMINI) || defined(SOLARIS)
# include <sys/mkdev.h>
#endif
#include <errno.h>
#include "devmap.h"


/* Map of major device numbers to ioctl functions */
ioctl_fnp	ioctlmap[NMAJ];


#define DEVMAP_NFDS	256	/* This 256 is just a guess. */

/* Cache of file descriptors and maj/min numbers */
dev_t		devmap_cache[DEVMAP_NFDS];
fd_set		devmap_valid;


/*
** Add a new ioctl handler for a particular major number to the device map.
*/
void
devmap_add_ioctl_handler(ioctl_fnp func, int maj)
{
	if(maj >= 0 && maj < NMAJ)
		ioctlmap[maj]=func;

	/* Invalidate devmap cache. */
	/* (This doesn't really belong here, but oh well.  It works.) */
	FD_ZERO(&devmap_valid);
}



/*
** Given the filename of a device node, return its major device number.
*/
int
devmap_major(const char *fn)
{
	struct stat st;

	if(stat(fn, &st) < 0)
		return -1;

	return major(st.st_rdev);
}


/*
** Find the appropriate ioctl handler and call it.
*/
int
lx_ioctl(int fd, int lx_cmd, unsigned long lx_arg)
{
	struct stat st;
	int maj, min;

	if(fd < 0)
		{
		errno=EBADF;
		return -1;
		}

	/*
	** If this causes problems, increase DEVMAP_NFDS and
	** recompile.  (And E-Mail the lxrun source maintainers
	** and tell them that DEVMAP_NFDS was exceeded!)
	*/
	if(fd >= DEVMAP_NFDS)
		{
		/*
		** Note that the fd might actually be valid, though more
		** likely, we just got passed a bogus fd.
		*/
		errno=EBADF;
		return -1;
		}

	if(!FD_ISSET(fd, &devmap_valid))
		{
		/* Maj/Min numbers are not in the cache.  Get them. */

		if(fstat(fd, &st) < 0)
			{
			errno=EBADF;
			return -1;
			}

		devmap_cache[fd] = st.st_rdev;

		FD_SET(fd, &devmap_valid);
		}

	maj=major(devmap_cache[fd]);
	min=minor(devmap_cache[fd]);

	return (ioctlmap[maj])(fd, lx_cmd, lx_arg, maj, min);
}


/*
** Invalidate a file descriptor in the cache.
** (This should be called by lx_close.)
*/
void
devmap_invalidate(int fd)
{
	if (fd >= DEVMAP_NFDS) {
		/* might be a broken program - just don't segv */
		return;
	}
	if (fd >= 0) {
		FD_CLR(fd, &devmap_valid);
	}
}
