/*
 * Copyright (C) 1996 Michael Davidson.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#pragma ident	"@(#)process.c	1.10	99/02/24 SMI"

#include	<errno.h>
#include	<fcntl.h>
#include	<limits.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include	<sys/mman.h>
#include	<sys/time.h>
#include	<sys/resource.h>
#include	"lxrun.h"
#include	"linux/types.h"
#include	"linux/resource.h"


/*
 * process control system calls
 */

/*
 *	execve
 *	ptrace		(NOT IMPLEMENTED)
 *	brk
 *	setrlimit
 *	getrlimit
 *	getrusage	(NOT IMPLEMENTED)
 *	getgroups
 *	setgroups
 *	getpriority	(NOT IMPLEMENTED)
 *	setpriority	(NOT IMPLEMENTED)
 *	uselib
 *	personality
 *      sched_setparam          (NOT IMPLEMENTED)
 *      sched_getparam          (NOT IMPLEMENTED)
 *      sched_setscheduler      (NOT IMPLEMENTED)
 *      sched_getscheduler      (NOT IMPLEMENTED)
 *      sched_yield             (NOT IMPLEMENTED)
 *      sched_get_priority_max  (NOT IMPLEMENTED)
 *      sched_get_priority_min  (NOT IMPLEMENTED)
 *      sched_rr_get_interval   (NOT IMPLEMENTED)
 */

/*
 * Nuke some environment variables. In particular, netscape likes to
 * LD_PRELOAD some linux libraries (/lib/libBrokenLocale.so.1) before
 * execing the netscape binary, but that invokes lxrun, and our ld doesn't
 * run lxrun because it can't find those libraries.
 */
static struct envtable {
	char	*var;
	size_t	len;
} envtable [] = {
	{ "LD_PRELOAD=", sizeof ("LD_PRELOAD=") - 1},
	{ NULL, 0 }
};

static void
clear_ld_environment(char **envp)
{
	struct envtable *etbl;
	for (; *envp != NULL; envp++) {
		for (etbl = &envtable[0]; etbl->len != 0; etbl++) {
			if (strncmp(*envp, etbl->var, etbl->len) == 0)
				(*envp)[etbl->len] = '\0';
		}
	}
}

int
lx_execve(char *prog, char **argv, char **envp)
{
	int ac = 0;
	char **av;
	extern char emulator[];
	char real_interp[PATH_MAX];

	clear_ld_environment(envp);
	if (emulator[0] == '\0') {
		if (is_shell_script(prog, real_interp, sizeof (real_interp))) {

			while (argv[ac++])
				continue;

			av =  alloca((ac + 1) * sizeof(char *));
			av[0] = argv[0];
			av[1] = prog;
			memcpy(&av[2], &argv[1], (ac - 1) * sizeof (char *));

			prog = real_interp;
			argv = av;
		}

	} else if (is_linux_exec(prog))  {
		/*
		 * Count args, including terminating NULL.
		 */
		while (argv[ac++])
			continue;

		/*
		 * Allocate new argv and copy old args.
		 * Use the "-0" flag to pass in the desired argv[0].
		 */
		av = alloca((ac + 3) * sizeof(char *));
		av[0] = emulator;
		av[1] = "-0";
		av[2] = argv[0];
		av[3] = prog;
		memcpy(&av[4], &argv[1], (ac - 1) * sizeof (char *));

		prog = emulator;
		argv = av;

	} else if (is_shell_script(prog, real_interp, sizeof (real_interp))) {

		while (argv[ac++])
			continue;

		/*
		 * Feed the script to the interpreter.
		 */
		av = alloca((ac + 4) * sizeof(char *));
		av[0] = emulator;
		av[1] = "-0";
		av[2] = argv[0];
		av[3] = real_interp;
		av[4] = prog;
		memcpy(&av[5], &argv[1], (ac - 1) * sizeof (char *));

		prog = emulator;
		argv = av;
	}

	return execve(prog, argv, envp);
}


int lx_ptrace()		{ return enosys("ptrace"); }

void *
lx_brk(caddr_t brk)
{
	unsigned long	new_brk;
	unsigned long	old_brk;
	int		z;
	extern caddr_t	minimum_brk;
	extern caddr_t	current_brk;

	if (brk < minimum_brk)
		return current_brk;

	new_brk = PAGE_ALIGN(brk);
	old_brk	= PAGE_ALIGN(current_brk);

	if (new_brk == old_brk)
		return current_brk = brk;

	if (new_brk < old_brk)
	{
		munmap((caddr_t)new_brk, old_brk - new_brk);
		return current_brk = brk;
	}
	
	if (dz_mmap((caddr_t)old_brk, new_brk - old_brk,
	    PROT_READ|PROT_WRITE|PROT_EXEC,
	    MAP_FIXED|MAP_PRIVATE, 0) == (caddr_t)old_brk)
	{
		current_brk = brk;
	}

	return current_brk;
}


int lx_setrlimit(int resource, const struct rlimit *rlp)	/* S.G. */
{
int rsrc;

switch(resource)
	{
	/* Cases that map directly from Linux to SCO */
	case LX_RLIMIT_CPU:	rsrc = RLIMIT_CPU; break;
	case LX_RLIMIT_FSIZE:	rsrc = RLIMIT_FSIZE; break;
	case LX_RLIMIT_DATA:	rsrc = RLIMIT_DATA; break;
	case LX_RLIMIT_STACK:	rsrc = RLIMIT_STACK; break;
	case LX_RLIMIT_CORE:	rsrc = RLIMIT_CORE; break;
	case LX_RLIMIT_NOFILE:	rsrc = RLIMIT_NOFILE; break;

	/* Cases that don't.  :-(  */
	case LX_RLIMIT_RSS:
		return enosys("setrlimit(RLIMIT_RSS)");
	case LX_RLIMIT_NPROC:
		return enosys("setrlimit(RLIMIT_NPROC)");
	default:
		return enosys("setrlimit(?)");
	}

return setrlimit(rsrc, rlp);
}

int lx_getrlimit(int resource, struct rlimit *rlp)	/* S.G. */
{
int rsrc;

switch(resource)
	{
	/* Cases that map directly from Linux to SCO */
	case LX_RLIMIT_CPU:	rsrc = RLIMIT_CPU; break;
	case LX_RLIMIT_FSIZE:	rsrc = RLIMIT_FSIZE; break;
	case LX_RLIMIT_DATA:	rsrc = RLIMIT_DATA; break;
	case LX_RLIMIT_STACK:	rsrc = RLIMIT_STACK; break;
	case LX_RLIMIT_CORE:	rsrc = RLIMIT_CORE; break;
	case LX_RLIMIT_NOFILE:	rsrc = RLIMIT_NOFILE; break;

	/* Cases that don't.  :-(  */
	case LX_RLIMIT_RSS:
		return enosys("getrlimit(RLIMIT_RSS)");
	case LX_RLIMIT_NPROC:
		return enosys("getrlimit(RLIMIT_NPROC)");
	default:
		return enosys("getrlimit(?)");
	}

return getrlimit(rsrc, rlp);
}

int lx_getrusage()	{ return enosys("getrusage"); }

int
lx_getgroups(int gidsetsize, lx_gid_t *lx_grouplist)
{
	gid_t	grouplist[NGROUPS_MAX];
	int	r;

	if (gidsetsize < 0 || gidsetsize > NGROUPS_MAX)
	{
		errno = EINVAL;
		return -1;
	}


	r = getgroups(gidsetsize, grouplist);

	if (r > 0 && gidsetsize > 0)
	{
		int	i;

		if (! verify_write(lx_grouplist, r * sizeof(lx_grouplist[0])))
			return -1;

		for (i = 0; i < r; i++)
			lx_grouplist[i] = (lx_gid_t) grouplist[i];
	}

	return r;
}	


int
lx_setgroups(int gidsetsize, lx_gid_t *lx_grouplist)
{
	gid_t	grouplist[NGROUPS_MAX];
	int	i;

	if (gidsetsize < 0 || gidsetsize > NGROUPS_MAX)
	{
		errno = EINVAL;
		return -1;
	}

	if (! verify_read(lx_grouplist, gidsetsize * sizeof(lx_grouplist[0])))
		return -1;

	for (i = 0; i < gidsetsize; i++)
		grouplist[i] = (gid_t) lx_grouplist[i];

	return setgroups(gidsetsize, grouplist);
}


int lx_getpriority()	{ return enosys("getpriority"); }
int lx_setpriority()	{ return enosys("setpriority"); }

int
lx_uselib(char *libname)
{
	if (load_library(libname) != (void *) -1)
		return 0;

	return -1;
}	


int
lx_personality(unsigned long personality)
{
	if (personality == 0xffffffff || personality == 0)
		return 0;
		
	errno = EINVAL;
	return -1;
}

int lx_sched_setparam()		{ return enosys("sched_setparam"); }
int lx_sched_getparam()		{ return enosys("sched_getparam"); }
int lx_sched_setscheduler()	{ return enosys("sched_setscheduler"); }
int lx_sched_getscheduler()	{ return enosys("sched_getscheduler"); }
int lx_sched_yield()		{ return enosys("sched_yield"); }
int lx_sched_get_priority_max()	{ return enosys("sched_get_priority_max"); }
int lx_sched_get_priority_min()	{ return enosys("sched_get_priority_min"); }
int lx_sched_rr_get_interval()	{ return enosys("sched_rr_get_interval"); }
